// Copyright (C) 2005, 2006 International Business Machines and others.
// All Rights Reserved.
// This code is published under the Common Public License.
//
// $Id: genome_ipopt_nlp.cpp,v 1.5 2010/05/12 03:05:11 andrones Exp $
//
// Authors:  Carl Laird, Andreas Waechter     IBM    2005-08-16

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cassert>

using namespace std;

#include "genome_ipopt_nlp.hpp"
#include "genome_model.hpp"
#include "coords.h"

// for printf
#ifdef HAVE_CSTDIO
# include <cstdio>
#else
# ifdef HAVE_STDIO_H
#  include <stdio.h>
# else
#  error "don't have header file for stdio"
# endif
#endif

typedef struct coords COORDS_T;

using namespace Ipopt;

// constructor
genome_ipopt_nlp::genome_ipopt_nlp()
{}

// my additional constructor 
genome_ipopt_nlp::genome_ipopt_nlp (GENOME *mygenome, double min_clash_dist, double min_clash_dist_inter, char* output_pdb, double sphere_radius, int use_weights, int bp_per_locus, int rDNA_frequency_normalizer, int weight_of_inter, double weight_unseen)
{
    this->mygenome = mygenome;
    this->output_pdb = output_pdb;
    this->min_clash_dist = min_clash_dist;
    this->min_clash_dist_inter = min_clash_dist_inter;
    this->sphere_radius = sphere_radius;
    this->use_weights = use_weights;
    this->weight_of_inter = weight_of_inter;
    this->bp_per_locus = bp_per_locus;
    this->rDNA_frequency_normalizer = rDNA_frequency_normalizer;
    this->weight_unseen = weight_unseen;

    // reporting multiple objective functions separately
    // Diament & Tuller, May 2014
    obj_function = new double[MAX_OBJ];
    for (int i=0; i<10; i++)
    	obj_function[i] = 0;
    //this->usedelta = true;
          
    centromeres[0] =  151584;  
    centromeres[1] =  238325;
    centromeres[2] =  114499;
    centromeres[3] =  449819;
    centromeres[4] =  152103;
    centromeres[5] =  148622;
    centromeres[6] =  497042;
    centromeres[7] =  105698;
    centromeres[8] =  355742;
    centromeres[9] =  436418;
    centromeres[10] =  439889;
    centromeres[11] = 150946;
    centromeres[12] =  268149;
    centromeres[13] =  628877;
    centromeres[14] =  326703;
    centromeres[15] =  556070;
    
    // TODO: this is a mess, it's repeated from genome_model.cpp, should be in one place
    chrlen[0] = 230208;
    chrlen[1] = 813178;
    chrlen[2] = 316617;
    chrlen[3] = 1531918;
    chrlen[4] = 576869;
    chrlen[5] = 270148;
    chrlen[6] = 1090947;
    chrlen[7] = 562643;
    chrlen[8] = 439885;
    chrlen[9] = 745745;
    chrlen[10] = 666454;
    chrlen[11] = 1078175 + SIZE_RDNA * NUM_RDNA_TO_ADD;
    chrlen[12] = 924429;
    chrlen[13] = 784333;
    chrlen[14] = 1091289;
    chrlen[15] = 948062;    
    
    
    iteration = 0;
    
    // The number of variables: 3 coordinates for each locus
    num_loci = mygenome->get_num_loci();
        
    num_interactions = mygenome->get_num_interactions();                    
        
    num_variables = num_loci*3;
    //num_variables += num_interactions;    // if usedelta
    
    printf ("num variables = %d\n", num_variables);    
    
    // Number of constraints
    // first add the number of loci, for the constraint to be in a sphere
    // NOTE: if I don't add any sphere constraints, it will be in a cube 
    num_sphere_constraints = num_loci;
    
    // add the number of adjacent loci, for the break constraints
    num_break_constraints = mygenome->get_num_adjacent_loci();
    
    // add the number of non-adjacent loci, for the clash constraints
    // NOTE: If I remove the clash constraints, the figure looks just terrible
    num_clash_constraints = num_loci*(num_loci-1)/2;
    // I could remove the num_break_constraints, but it's simpler to leave them in
        
    // number of distance constraints = number of interactions
    if (use_weights == 2)
        num_dist_constraints = num_interactions;
    else
        num_dist_constraints = 0;
    
    // Constraints for the centromeric area and for the nucleolus
    // 16 for the centromeres of all chromosomes
    // and a bunch for the rDNA regions
    // Actually, I'm just constraining chrom 12 to be inside that area
    num_small_sphere_constraints = 1 + NUM_RDNA_TO_ADD * SIZE_RDNA / bp_per_locus;    
    
    
    num_constraints = num_sphere_constraints + num_break_constraints + num_clash_constraints + num_dist_constraints + num_small_sphere_constraints;
    //num_constraints += num_interactions;  // if usedelta 
    
    printf ("num constraints = %d = %d + %d + %d + %d + %d\n", 
        num_constraints, num_sphere_constraints, num_break_constraints, 
        num_clash_constraints, num_dist_constraints, num_small_sphere_constraints);
    
    min_dist = mygenome -> get_min_dist();
    max_dist = mygenome -> get_max_dist();
    
    // initialize to the given starting point
//     for (Index i=0; i < num_loci*3; i+=3)
//     {
//         COORDS_T *coord = mygenome->get_coords (i/3);
//         if (coord == NULL)
//             die ("Something wrong in GENOME::get_coords, returned NULL\n");
//         printf ("%g %g %g\n", (coord->x+1)*100, (coord->y+1)*100, (coord->z+1)*100);
//     }
    
    //printf ("Constructor OK\n");
    
     
    
}


//destructor
genome_ipopt_nlp::~genome_ipopt_nlp()
{}


// returns the size of the problem
bool genome_ipopt_nlp::get_nlp_info(Index& n, Index& m, Index& nnz_jac_g,
                             Index& nnz_h_lag, IndexStyleEnum& index_style)
{
    //printf ("Trying NLP info\n");
    // The number of variables
    n = num_variables;
    
    // Number of constraints    
    m = num_constraints; 
    
    // The number of non-zero elements in the Jacobian
    nnz_jac_g = num_sphere_constraints * 3 +
                num_break_constraints * 6 +
                num_clash_constraints * 6 + 
                num_dist_constraints * 6 + 
                num_small_sphere_constraints * 3;
    //nnz_jac_g += num_interactions * 7;    // if usedelta
    
    // number of non-zero elements in the Hessian
    // don't need to set nnz_h_lag if I use quasi-newton
    
    // use the C style indexing (0-based)
    index_style = TNLP::C_STYLE;
    
    //printf ("NLP info OK\n");
    
    return true;
}

// returns the variable bounds
bool genome_ipopt_nlp::get_bounds_info(Index n, Number* x_l, Number* x_u,
                                Index m, Number* g_l, Number* g_u)
{
    //printf ("Trying get bounds\n");
    // here, the n and m we gave IPOPT in get_nlp_info are passed back to us.
    // If desired, we could assert to make sure they are what we think they are.
    assert(n == num_variables);
    assert(m == num_constraints);
    
    // the bounds on the variables
    
    // first the xyz locations
    // we are in a sphere with radius sphere_radius    
    for (Index i=0; i<num_loci*3; i++)
    {
        x_l[i] = -sphere_radius;
        x_u[i] = sphere_radius;
    }
    
    // if usedelta
    /*
    // next the distance for each interaction
    // this distance has to be more than min_interaction_dist, and less than the diameter of the sphere
    for (Index i=num_loci*3; i < n; i++)
    {        
        x_l[i] = min_interaction_dist;
        x_u[i] = 2;     // we are in a unit sphere
        
        //x_l[i] = -2e19;
        //x_u[i] = 2e19;     // we are in a unit sphere        
    }
    */
    
    // the bounds on the constraints        
    // no upper bound is 2e19
    
    // CAREFUL: I have to make sure I have the right order of constraints
    
    // first the sphere constraints
    // 0 <= d^2(p_i,0) <= 1, for all loci p_i
    for (Index i=0; i<num_sphere_constraints; i++)
    {
        g_l[i] = -2e19;     //0.0;      // not sure what's faster
        g_u[i] = sphere_radius*sphere_radius;
    }
    
    int msofar = num_sphere_constraints;
    
    for (Index i=msofar; i < msofar + num_break_constraints; i++)
    {
        // TODO: if I just set the lower bound to 0, is it easier to solve?
        // I tested this for resolution 1000, chr1, and it was 123 iterations with 0, and 139 iterations with min_dist. 
        g_l[i] = min_dist*min_dist;     // this is already covered by the clash constraints. I could just use -2e19, but it doesn't really speed it up
        g_u[i] = max_dist*max_dist;
    }
    
    msofar += num_break_constraints;
    
    Index i = msofar;
    for (int i1=0; i1 < num_loci; i1++)
    {
        int chr1 = mygenome->get_chromosome(i1);        
        for (int i2=i1+1; i2 < num_loci; i2++)
        {            
            int chr2 = mygenome->get_chromosome(i2);
            //printf ("Loci %d and %d, chroms %d and %d\t", i1, i2, chr1, chr2);            
            if (chr1 == chr2)   // intra
            {
                g_l[i] = min_clash_dist*min_clash_dist;
                //printf ("Lower bound: %g\n", min_clash_dist);
            }
            else    // inter
            {
                g_l[i] = min_clash_dist_inter*min_clash_dist_inter;
                //printf ("Lower bound: %g\n", min_clash_dist_inter);
            }            
            g_u[i] = 2e19;
            i++;
        }
    } 
/*    for (Index i=msofar; i < msofar + num_clash_constraints; i++)
    {        
        g_l[i] = min_clash_dist*min_clash_dist;
        g_u[i] = 2e19;      //sphere_radius*2*sphere_radius*2; // diameter is 2, and this is 2^2
        // It's ok not to set any upper bound here. That is satisfied because of the sphere constraints
    }            */
    
    msofar += num_clash_constraints;
    
    // the distance constraints
/*    if (num_dist_constraints > 0)
    {    
        for (Index i=msofar; i < msofar + num_dist_constraints; i++)
        {
            g_l[i] = 0.1*0.1;
            g_u[i] = sphere_radius*2*sphere_radius*2; // diameter is 2, and this is 2^2
        }
    }  */      
       
    // the distance constraints
    i=msofar;
    if (num_dist_constraints > 0)
    {         
        INTERACTION_T *interaction = mygenome->get_first_interaction();
        while (interaction != NULL)
        {
            //printf ("interaction frequency: %d\n", interaction->freq);
            g_l[i] = 1.0/interaction->freq * 1.0/interaction->freq - min_dist*min_dist;
            g_u[i] = sphere_radius*2*sphere_radius*2; // diameter is 2, and this is 2^2            
            i++;
            interaction = interaction->next;
        }
    }          
    msofar += num_dist_constraints;
         
    g_l[msofar] = -2e19;
    g_u[msofar] = CENTRO_RADIUS * CENTRO_RADIUS;  
    for (Index i=msofar+1; i < msofar + num_small_sphere_constraints; i++)
    {
        g_l[i] = -2e19;
        g_u[i] = NUCLEO_RADIUS * NUCLEO_RADIUS;
    }      
    msofar += num_small_sphere_constraints;
    
    
    // if usedelta
    /*
    // bounds of the interaction constraints
    for (Index i=msofar; i < msofar + num_interactions; i++)
    {
        g_l[i] = 0;
        g_u[i] = 0;
    }
    msofar += num_interactions;
    */
    
    assert (m == msofar);
    
    //printf ("BOUNDS info OK\n");
    return true;
}


// returns the initial point for the problem
bool genome_ipopt_nlp::get_starting_point(Index n, bool init_x, Number* x,
                                   bool init_z, Number* z_L, Number* z_U,
                                   Index m, bool init_lambda,
                                   Number* lambda)
{
    // Here, we assume we only have starting values for x, if you code
    // your own NLP, you can provide starting values for the dual variables
    // if you wish
    
    //printf ("Trying starting point ...\n");
    assert(init_x == true);
    assert(init_z == false);
    assert(init_lambda == false);
    
    assert (n==num_variables);
    
    
    // initialize to the given starting point
    for (Index i=0; i < num_loci*3; i+=3)
    {
        COORDS_T *coord = mygenome->get_coords (i/3);
        if (coord == NULL)
            die ("Something wrong in GENOME::get_coords, returned NULL\n");
        // NOTE: I tried to set these coordinates to 0.0 or 0.9. The IPOPT problem fails (it says Restoration failed)             
        x[i]   = coord->x;  // x coordinate
        x[i+1] = coord->y;  // y coordinate
        x[i+2] = coord->z;  // z coordinate
        //printf ("Coordinates: %g %g %g\n", coord->x, coord->y, coord->z);
        //fflush(stdout);
    }
    
    // if usedelta
    /*
    // the interaction distance    
    for (Index i = num_loci*3; i < n; i++)
    {
        x[i] = min_interaction_dist;
    }        
    
    assert (n==num_loci*3+num_interactions);
    */
    
    //printf ("STArting point OK\n");
    return true;
}

int genome_ipopt_nlp::is_inter_arms (INTERACTION_T *interaction)
// return 1 if this interaction is between 2 arms (of the same or different chrom)
// return 0 is this is intra within the same arm
{
    if (interaction->chrom1_index != interaction->chrom2_index)
        return 1;
          
    // same chromosomes          
    int centro_locus = (int)(centromeres[interaction->chrom1_index]/bp_per_locus);
    if (( interaction->locus1_index < centro_locus && interaction->locus2_index > centro_locus) ||
        ( interaction->locus1_index > centro_locus && interaction->locus2_index < centro_locus)
        )
       return 1;
           
    return 0;        
}

int genome_ipopt_nlp::is_inter_centro_telo (INTERACTION_T *interaction)
// return 1 if this interaction is between the centromere area and the telomere area of the same chromosome
// return 0 if it's not
{
    if (interaction->chrom1_index != interaction->chrom2_index)
        return 0;
          
    // same chromosomes          
    int centro_locus = (int)(centromeres[interaction->chrom1_index]/bp_per_locus);
    
    // NOTE: I'm not using the telomere information here. TODO???
    int telo_locus = (int)(chrlen[interaction->chrom1_index]/bp_per_locus);
        
    int centro1 = centro_locus - WINDOW/bp_per_locus;
    int centro2 = centro_locus + WINDOW/bp_per_locus;
    int telo1 = 2*WINDOW/bp_per_locus;
    int telo2 = telo_locus - 2*WINDOW/bp_per_locus;
    
    printf ("Is inter centro telo: %d %d %d %d -- %d - %d %d %d - %d\n",
            interaction->chrom1_index, interaction->locus1_index,
            interaction->chrom2_index, interaction->locus2_index,
            telo1, centro1, centro_locus, centro2, telo2);    
    
    
    // left arm
    if (interaction->locus1_index >= centro1 && 
        interaction->locus1_index <= centro_locus && 
        interaction->locus2_index <= telo1)
        return 1;
        
    // reverse of left arm
    if (interaction->locus2_index >= centro1 && 
        interaction->locus2_index <= centro_locus && 
        interaction->locus1_index <= telo1)
        return 1;   
                 
    // right arm                 
    if (interaction->locus1_index >= centro_locus && 
        interaction->locus1_index <= centro2 && 
        interaction->locus2_index >= telo2)
        return 1;     
               
    //reverse of the right arm
    if (interaction->locus2_index >= centro_locus && 
        interaction->locus2_index <= centro2 && 
        interaction->locus1_index >= telo2)
        return 1;     
           
    return 0;        
}



// returns the value of the objective function
bool genome_ipopt_nlp::eval_f(Index n, const Number* x, bool new_x, Number& obj_value)
{
    //printf ("Trying eval f\n");
    assert(n == num_variables);

    // reporting multiple objective functions separately
    // Diament & Tuller, May 2014
    obj_value = 0;
    for (int i=0; i<10; i++)
    	obj_function[i] = 0;
    
    // if usedelta
    /*
    for (Index i = num_loci*3; i < num_variables; i++)
    {
        obj_value += x[i]*x[i];
    }
    */    
    if (num_interactions > 0)
    {                      
        INTERACTION_T *interaction = mygenome->get_first_interaction();
        while (interaction != NULL)
        {
            int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
            int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);
            if (use_weights == 5 || use_weights == 6)   // use wish_dist
            {
                // Unfortunately, this function won't be quadratic any more
                double d   = sqrt( (x[l]-x[r])*(x[l]-x[r]) + 
                                   (x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                                   (x[l+2]-x[r+2])*(x[l+2]-x[r+2])) - 
                             interaction->wish_dist;
                             
                // add the weight_of_intra, if applicable
                if (is_inter_arms(interaction))
                {
                    d = d*weight_of_inter;
                }
                
//                 if (is_inter_centro_telo(interaction))
//                 {
//                     printf ("Is inter centro telo: %d %d %d %d\n",
//                         interaction->chrom1_index, interaction->locus1_index,
//                         interaction->chrom2_index, interaction->locus2_index);
//                 }
//                 else
//                     printf ("NOT INTER centro telo\n");

                // reporting multiple objective functions separately
                // Diament & Tuller, May 2014
                obj_value += d*d;
                obj_function[interaction->priority] += d*d;
                //obj_function[interaction->priority] += d*d;
            }
            else    // use weights in the objective function 
            {                        
                double w;
                switch (use_weights)
                {
                    case 0: w = 1; break;   
                    case 1: w = interaction->freq; break;
                    case 2: w = 1; break;
                    case 3: w = interaction->inversep; break;
                    case 4: w = interaction->inverseq; break;
                    default: w = 1;
                }     
            // printf ("%d %d = %d ; %d %d = %d\n", 
            //         interaction->chrom1_index, interaction->locus1_index, l,
            //         interaction->chrom2_index, interaction->locus2_index, r);
                
                // NOTE: f can be negative (for example for centromere of 12 vs rDNA), so I have to keep the sign
                if (w > 0)
                {
                    obj_value += 
                        w*w*(x[l]-x[r])*(x[l]-x[r]) + 
                        w*w*(x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                        w*w*(x[l+2]-x[r+2])*(x[l+2]-x[r+2]);
                }
                else
                {
                    obj_value = obj_value - 
                        w*w*(x[l]-x[r])*(x[l]-x[r]) - 
                        w*w*(x[l+1]-x[r+1])*(x[l+1]-x[r+1]) - 
                        w*w*(x[l+2]-x[r+2])*(x[l+2]-x[r+2]);               
                }
            }
            interaction = interaction->next;
        }
    }        
        
    // add the interactions with the center of the nucleolus
    int num_rDNA_interactions = mygenome->get_num_rDNA_interactions();    
    if (num_rDNA_interactions > 0)
    {                      
        INTERACTION_T *interaction = mygenome->get_first_rDNA_interaction();
        while (interaction != NULL)
        {
            int l;
            if (interaction->chrom1_index == 11)
                l = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);
            else if (interaction->chrom2_index == 11)
                l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
            else
            {
                printf ("One of the 2 ends has to be in chromosome 12! ABORT!\n");
                exit(1);
            }
            double w = interaction->freq/rDNA_frequency_normalizer;
            // I'm keeping the full frequency because all the repeats are in the nucleolus
            
            obj_value += 
                w*w*(x[l]-NUCLEO_RADIUS)*(x[l]-NUCLEO_RADIUS) + 
                w*w* x[l+1] * x[l+1] + 
                w*w* x[l+2] * x[l+2];
            interaction = interaction->next;
        }
    }             
        
    // try to add unseen interactions, so that I don't put things closer to each other than needed        
    int num_unseen_interactions = mygenome->get_num_unseen_interactions();    
    if (num_unseen_interactions > 0)
    {                      
        INTERACTION_T *interaction = mygenome->get_first_unseen_interaction();
        while (interaction != NULL)
        {
            int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
            int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);
            if (use_weights == 5 || use_weights == 6)   // use wish_dist
            {
                // Unfortunately, this function won't be quadratic any more
                double d   = sqrt( (x[l]-x[r])*(x[l]-x[r]) + 
                                   (x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                                   (x[l+2]-x[r+2])*(x[l+2]-x[r+2])) - 
                             interaction->wish_dist;
                // multiply by the weight_unseen
                d = d*weight_unseen;                                                          
                obj_value += d*d;                                    
            }            
            interaction = interaction->next;
        }
    }             
            
    
    // Save the file every 10 iterations, so that we can check what it's doin'
    iteration++;
    if (iteration%10 == 0)
    {
        mygenome->set_coords((double*)x);
        char temp_filename[1000];
        sprintf (temp_filename, "%s.temp.pdb", output_pdb); 
        mygenome->print_1D_3D_genome(temp_filename); // Diament & Tuller, May 2014
        sprintf (temp_filename, "%s.temp.txt", output_pdb); 
        mygenome->save_txt(temp_filename);
        // reporting multiple objective functions separately
        // Diament & Tuller, May 2014
        printf( "objective\t");
        for (int i=0; i<MAX_OBJ; i++)
        {
        	if (obj_function[i] != 0)
        		printf( "%.5e\t", obj_function[i]);
        }
        printf("\n");
    }
    return true;
}

// return the gradient of the objective function grad_{x} f(x)
bool genome_ipopt_nlp::eval_grad_f(Index n, const Number* x, bool new_x, Number* grad_f)
{
    //printf ("Trying eval grad f\n");
    assert(n == num_variables);
        
    for (Index i=0; i<num_loci*3; i++)
    {
        grad_f[i] = 0;
    }
    
    
    // if usedelta
    /*
    for (Index i = num_loci*3; i < num_variables; i++)
    {
        grad_f[i] = 2*x[i];
    }
    */
    
    if (num_interactions > 0)
    {                      
        INTERACTION_T *interaction = mygenome->get_first_interaction();
        while (interaction != NULL)
        {
            //printf ("%d, %d\t", interaction->chrom1_index, interaction->locus1_index);
            //printf ("%d, %d\n", interaction->chrom2_index, interaction->locus2_index);
            int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
            int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);
            if (use_weights == 5 || use_weights == 6)   // use wish_dist
            {
                // Unfortunately, this function won't be quadratic any more
                double sq = sqrt( (x[l]-x[r])*(x[l]-x[r]) + 
                                   (x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                                   (x[l+2]-x[r+2])*(x[l+2]-x[r+2]));
                double di = sq - interaction->wish_dist;
                
                // add the weight_of_inter, if applicable
                
                if (is_inter_arms(interaction))
                {
                    di = di*weight_of_inter;
                }
                
                // NOT SURE why this is this way, but it looks wrong to me (Dec 2, 2009)
				// No, this is right, because there's a square
                grad_f[l]   += di / sq * 2*(x[l]-x[r]);
                grad_f[l+1] += di / sq * 2*(x[l+1]-x[r+1]);
                grad_f[l+2] += di / sq * 2*(x[l+2]-x[r+2]);
                grad_f[r]   -= di / sq * 2*(x[l]-x[r]);
                grad_f[r+1] -= di / sq * 2*(x[l+1]-x[r+1]); 
                grad_f[r+2] -= di / sq * 2*(x[l+2]-x[r+2]);
            }
            else 
            {            
                double w;
                switch (use_weights)
                {
                    case 0: w = 1; break;   
                    case 1: w = interaction->freq; break;
                    case 2: w = 1; break;
                    case 3: w = interaction->inversep; break;
                    case 4: w = interaction->inverseq; break;
                    default: w = 1;
                }  
            // printf ("%d %d = %d ; %d %d = %d\n", 
            //         interaction->chrom1_index, interaction->locus1_index, l,
            //         interaction->chrom2_index, interaction->locus2_index, r);
                
                if (w > 0)
                {
                    grad_f[l] += 2*w*w*(x[l]-x[r]);
                    grad_f[l+1] += 2*w*w*(x[l+1]-x[r+1]);
                    grad_f[l+2] += 2*w*w*(x[l+2]-x[r+2]);
                    grad_f[r] += -2*w*w*(x[l]-x[r]);
                    grad_f[r+1] += -2*w*w*(x[l+1]-x[r+1]);
                    grad_f[r+2] += -2*w*w*(x[l+2]-x[r+2]);
                }
                else
                {
                    grad_f[l] -= 2*w*w*(x[l]-x[r]);
                    grad_f[l+1] -= 2*w*w*(x[l+1]-x[r+1]);
                    grad_f[l+2] -= 2*w*w*(x[l+2]-x[r+2]);
                    grad_f[r] -= -2*w*w*(x[l]-x[r]);
                    grad_f[r+1] -= -2*w*w*(x[l+1]-x[r+1]);
                    grad_f[r+2] -= -2*w*w*(x[l+2]-x[r+2]);
                }
            }        
            interaction = interaction->next;
        }
    }  
    
    // add the interactions with the center of the nucleolus
    int num_rDNA_interactions = mygenome->get_num_rDNA_interactions();    
    if (num_rDNA_interactions > 0)
    {                      
        INTERACTION_T *interaction = mygenome->get_first_rDNA_interaction();
        while (interaction != NULL)
        {
            int l;
            if (interaction->chrom1_index == 11)
                l = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);
            else if (interaction->chrom2_index == 11)
                l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
            else
            {
                printf ("One of the 2 ends has to be in chromosome 12! ABORT!\n");
                exit(1);
            }
            double w = interaction->freq/rDNA_frequency_normalizer;
            // I'm keeping the full frequency because all the repeats are in the nucleolus
            
            grad_f[l] += 2*w*w*(x[l]-NUCLEO_RADIUS);
            grad_f[l+1] += 2*w*w* x[l+1];
            grad_f[l+2] += 2*w*w* x[l+2];
            interaction = interaction->next;
        }
    }       
    
    // try to add unseen interactions, so that I don't put things closer to each other than needed        
    int num_unseen_interactions = mygenome->get_num_unseen_interactions();    
    if (num_unseen_interactions > 0)
    {                      
        INTERACTION_T *interaction = mygenome->get_first_unseen_interaction();
        while (interaction != NULL)
        {
            int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
            int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);
            if (use_weights == 5 || use_weights == 6)   // use wish_dist
            {
                // Unfortunately, this function won't be quadratic any more
                double sq = sqrt( (x[l]-x[r])*(x[l]-x[r]) + 
                                   (x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                                   (x[l+2]-x[r+2])*(x[l+2]-x[r+2]));
                double di = sq - interaction->wish_dist;
                
                // add the weight_of_inter, if applicable                
                di = di*weight_unseen;
                grad_f[l]   += di / sq * 2*(x[l]-x[r]);
                grad_f[l+1] += di / sq * 2*(x[l+1]-x[r+1]);
                grad_f[l+2] += di / sq * 2*(x[l+2]-x[r+2]);
                grad_f[r]   -= di / sq * 2*(x[l]-x[r]);
                grad_f[r+1] -= di / sq * 2*(x[l+1]-x[r+1]); 
                grad_f[r+2] -= di / sq * 2*(x[l+2]-x[r+2]);
            }                      
            interaction = interaction->next;
        }
    }      
    
    
    //printf ("Eval grad f OK\n");    
    return true;
}

// return the value of the constraints: g(x)
bool genome_ipopt_nlp::eval_g(Index n, const Number* x, bool new_x, Index m, Number* g)
{
    //printf ("Trying eval g\n");
    assert(n == num_variables);
    assert(m == num_constraints);
    
    int xvari = num_loci*3;
    
    // CAREFUL: I have to make sure I have the right order of constraints
    
    // first the sphere constraints
    // 0 <= d^2(p_i,0) <= 1, for all loci p_i
    for (Index i=0; i<num_sphere_constraints*3; i+=3)
    {
        g[i/3] = x[i]*x[i] + x[i+1]*x[i+1] + x[i+2]*x[i+2];
    }
    
    // the break constraints
    Index consi = num_sphere_constraints;    
    int i=0; 
    if (num_break_constraints > 0)
    {
        for (int c=0; c < mygenome->get_num_chroms(); c++)
        {
            for (int l=0; l < mygenome->get_num_loci(c)-1; l++)
            {        
                g[consi++] = 
                    (x[i]-x[i+3])*(x[i]-x[i+3]) + (x[i+1]-x[i+4])*(x[i+1]-x[i+4]) + (x[i+2]-x[i+5])*(x[i+2]-x[i+5]);
                i+=3;
            }
            i+=3;
        }
    }
    
    //vari += num_break_constraints;
    // the clash constraints
    assert (consi == num_sphere_constraints + num_break_constraints);
    if (num_clash_constraints > 0)
    {
        int j;        
        for (i=0; i < (num_loci-1)*3; i+=3)
        {
            for (j=i+3; j < num_loci*3; j+=3)
            {
                g[consi++] = (x[i]-x[j])*(x[i]-x[j]) + (x[i+1]-x[j+1])*(x[i+1]-x[j+1]) + (x[i+2]-x[j+2])*(x[i+2]-x[j+2]);
            }
        }                        
    }
    
    assert (consi == num_sphere_constraints + num_break_constraints + num_clash_constraints);
    
    // the distance constraints
    if (num_dist_constraints > 0)
    {         
        INTERACTION_T *interaction = mygenome->get_first_interaction();
        while (interaction != NULL)
        {
            int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
            int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);;
           // printf ("%d %d = %d ; %d %d = %d\n", 
           //         interaction->chrom1_index, interaction->locus1_index, l,
           //         interaction->chrom2_index, interaction->locus2_index, r);
            
            g[consi] = 
                (x[l]-x[r])*(x[l]-x[r]) + 
                (x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                (x[l+2]-x[r+2])*(x[l+2]-x[r+2]);
            consi++;
            interaction = interaction->next;
        }
    }           
    
    // the constraints for centromere area and nucleolus
    // add centromere of chromosome 12  150946, and all others
/*     for (int i=1; i <= NUM_CHROMS; i++)
     {
         int chr = mygenome->get_chrom_index(i);
         int centro_x = mygenome->get_x_index(chr, centromeres[chr]/bp_per_locus); 
         g[consi] =  (x[centro_x]-CENTRO_CENTER)*(x[centro_x]-CENTRO_CENTER) + 
                     x[centro_x+1]*x[centro_x+1] +
                     x[centro_x+2]*x[centro_x+2]; 
         consi++;
     }        
*/  
    // No, only add centromere 12
    int chr = mygenome->get_chrom_index(12);
    int centro_x = mygenome->get_x_index(chr, centromeres[chr]/bp_per_locus); 
    g[consi] =  (x[centro_x]-CENTRO_CENTER)*(x[centro_x]-CENTRO_CENTER) + 
                x[centro_x+1]*x[centro_x+1] +
                x[centro_x+2]*x[centro_x+2]; 
    consi++;
    
    for (Index i=0; i<num_small_sphere_constraints-1; i++)
    {
        int chr12 = mygenome->get_chrom_index(12);
        int centro_x = mygenome->get_x_index(chr12, RDNA_POSITION/bp_per_locus + i);
        g[consi] = (x[centro_x]-NUCLEO_CENTER)*(x[centro_x]-NUCLEO_CENTER) + 
                x[centro_x+1]*x[centro_x+1] +
                x[centro_x+2]*x[centro_x+2]; 
        consi++;                
    }

  
    // the interaction constraints  
      
    // if usedelta
    /*  
    if (num_interactions > 0)
    {
        // CAUTIOUS!! the index of the constraints (vari) is different from the index of the x variable (xvari)                
        INTERACTION_T *interaction = mygenome->get_first_interaction();
        while (interaction != NULL)
        {
            int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
            int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);;
           // printf ("%d %d = %d ; %d %d = %d\n", 
           //         interaction->chrom1_index, interaction->locus1_index, l,
           //         interaction->chrom2_index, interaction->locus2_index, r);
            
            g[consi] = 
                (x[l]-x[r])*(x[l]-x[r]) + 
                (x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                (x[l+2]-x[r+2])*(x[l+2]-x[r+2]) - x[xvari]*x[xvari];
            consi++;
            xvari++;
            interaction = interaction->next;
        }
    }        
    */
    assert (m == consi);
    assert (n == xvari);
       
    
    //printf ("Eval g OK\n");
    return true;
}

// return the structure or values of the jacobian
bool genome_ipopt_nlp::eval_jac_g(Index n, const Number* x, bool new_x,
                           Index m, Index nele_jac, Index* iRow, Index *jCol,
                           Number* values)
{
    //printf ("Trying jac g\n");
    int ijac = 0;
    int xvari = num_loci*3;
    if (values == NULL) {
        // return the structure of the jacobian
    
        // CAREFUL: I have to make sure I have the right order of constraints
        
        // first the sphere constraints
        // 0 <= d^2(p_i,0) <= 1, for all loci p_i
        for (Index i=0; i<num_sphere_constraints*3; i+=3)
        {
            iRow[ijac] = i/3;
            jCol[ijac++] = i;            
             
            iRow[ijac] = i/3;
            jCol[ijac++] = i+1;            
            
            iRow[ijac] = i/3;
            jCol[ijac++] = i+2;                    
        }    
        
        Index consi = num_sphere_constraints;
        
        // the break constraints
        if (num_break_constraints > 0)
        {                
            int i=0; 
            for (int c=0; c < mygenome->get_num_chroms(); c++)
            {
                for (int l=0; l < mygenome->get_num_loci(c)-1; l++)
                {        
                    iRow[ijac] = consi;
                    jCol[ijac++] = i;
                    
                    iRow[ijac] = consi;
                    jCol[ijac++] = i+1;
                    
                    iRow[ijac] = consi;
                    jCol[ijac++] = i+2;
                    
                    iRow[ijac] = consi;
                    jCol[ijac++] = i+3;
                                    
                    iRow[ijac] = consi;
                    jCol[ijac++] = i+4;
                    
                    iRow[ijac] = consi;
                    jCol[ijac++] = i+5;
                                                    
                    consi++;                                
                    i+=3;
                    //g[consi++] = 
                    //    (x[i]-x[i+3])*(x[i]-x[i+3]) + (x[i+1]-x[i+4])*(x[i+1]-x[i+4]) + (x[i+2]-x[i+5])*(x[i+2]-x[i+5])                
                }
                i+=3;
            }
        }
        assert (consi == num_sphere_constraints + num_break_constraints);
        
        if (num_clash_constraints > 0)
        {
            // the clash constraints
            int j;
            //vari += num_break_constraints;
            for (int i=0; i < (num_loci-1)*3; i+=3)
            {
                for (j=i+3; j < num_loci*3; j+=3)
                {                
                    iRow[ijac] = consi;
                    jCol[ijac++] = i;
                    
                    iRow[ijac] = consi;
                    jCol[ijac++] = i+1;
                                    
                    iRow[ijac] = consi;
                    jCol[ijac++] = i+2;                                
                    
                    iRow[ijac] = consi;
                    jCol[ijac++] = j;
                    
                    iRow[ijac] = consi;
                    jCol[ijac++] = j+1;
                                    
                    iRow[ijac] = consi;
                    jCol[ijac++] = j+2;
                                    
                    consi++;                                    
                    //g[consi++] = (x[i]-x[j])*(x[i]-x[j]) + (x[i+1]-x[j+1])*(x[i+1]-x[j+1]) + (x[i+2]-x[j+2])*(x[i+2]-x[j+2]);
                }
            }
        }
        assert (consi == num_sphere_constraints + num_break_constraints + num_clash_constraints);        
        
        // the distance constraints
        if (num_dist_constraints > 0)
        {                                  
            INTERACTION_T *interaction = mygenome->get_first_interaction();
            while (interaction != NULL)
            {
                int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
                int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);;
                        
                //printf ("Begin int iRow/jCol: ijac = %d, consi = %d\n", ijac, consi);                        
                        
                iRow[ijac] = consi;
                jCol[ijac++] = l;
                
                iRow[ijac] = consi;
                jCol[ijac++] = l+1;
                            
                iRow[ijac] = consi;
                jCol[ijac++] = l+2;
                                        
                iRow[ijac] = consi;
                jCol[ijac++] = r;
                                                    
                iRow[ijac] = consi;
                jCol[ijac++] = r+1;
                                                                
                iRow[ijac] = consi;
                jCol[ijac++] = r+2;
                                                                                            
                //g[consi] = (x[l]-x[r])*(x[l]-x[r]) + 
                //          (x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                //          (x[l+2]-x[r+2])*(x[l+2]-x[r+2]) - x[xvari]*x[xvari];                            
                     
                interaction = interaction->next;
                consi++;
            }
        }
        
        for (Index i=consi; i<num_small_sphere_constraints*3; i+=3)
        {
            iRow[ijac] = i/3;
            jCol[ijac++] = i;            
             
            iRow[ijac] = i/3;
            jCol[ijac++] = i+1;            
            
            iRow[ijac] = i/3;
            jCol[ijac++] = i+2;                    
        }          
        
        //for (int i=1; i <= NUM_CHROMS; i++)
        {
            int i=12;
            int chr = mygenome->get_chrom_index(i);
            int centro_x = mygenome->get_x_index(chr, centromeres[chr]/bp_per_locus); 
                                    
            iRow[ijac] = consi;
            jCol[ijac++] = centro_x;            
                
            iRow[ijac] = consi;
            jCol[ijac++] = centro_x+1;            
            
            iRow[ijac] = consi;
            jCol[ijac++] = centro_x+2;                       
            consi++;
        }
              
        for (Index i=0; i<num_small_sphere_constraints-1; i++)
        {
            int chr12 = mygenome->get_chrom_index(12);
            int centro_x = mygenome->get_x_index(chr12, RDNA_POSITION/bp_per_locus + i);

            iRow[ijac] = consi;
            jCol[ijac++] = centro_x;            
                
            iRow[ijac] = consi;
            jCol[ijac++] = centro_x+1;            
            
            iRow[ijac] = consi;
            jCol[ijac++] = centro_x+2;  
        
            consi++;                
        }        
        
        // if usedelta
        /*
        if (num_interactions > 0)
        {                                  
            INTERACTION_T *interaction = mygenome->get_first_interaction();
            while (interaction != NULL)
            {
                int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
                int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);;
                        
                //printf ("Begin int iRow/jCol: ijac = %d, consi = %d\n", ijac, consi);                        
                        
                iRow[ijac] = consi;
                jCol[ijac++] = l;
                
                iRow[ijac] = consi;
                jCol[ijac++] = l+1;
                            
                iRow[ijac] = consi;
                jCol[ijac++] = l+2;
                                        
                iRow[ijac] = consi;
                jCol[ijac++] = r;
                                                    
                iRow[ijac] = consi;
                jCol[ijac++] = r+1;
                                                                
                iRow[ijac] = consi;
                jCol[ijac++] = r+2;
                                                                            
                iRow[ijac] = consi;
                jCol[ijac++] = xvari;
                
                //g[consi] = (x[l]-x[r])*(x[l]-x[r]) + 
                //          (x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                //          (x[l+2]-x[r+2])*(x[l+2]-x[r+2]) - x[xvari]*x[xvari];                            
                     
                interaction = interaction->next;
                consi++;
                xvari++;
            }
        }
        */
        
        assert (consi == m);
    }
    else {
        // return the values of the jacobian of the constraints
            
        for (Index i=0; i<num_sphere_constraints*3; i+=3)
        {
            values[ijac++] = 2*x[i];    // i/3, i
            values[ijac++] = 2*x[i+1];  // i/3, i+1
            values[ijac++] = 2*x[i+2];  // i/3, i+2
        }      
        
        Index consi = num_sphere_constraints;    
        if (num_break_constraints > 0)
        {
            // the break constraints            
            int i=0; 
            for (int c=0; c < mygenome->get_num_chroms(); c++)
            {
                for (int l=0; l < mygenome->get_num_loci(c)-1; l++)
                {        
                    values[ijac++] = 2*(x[i]-x[i+3]);   // consi, i
                    values[ijac++] = 2*(x[i+1]-x[i+4]);   // consi, i+1
                    values[ijac++] = 2*(x[i+2]-x[i+5]);   // consi, i+2
                    values[ijac++] = -2*(x[i]-x[i+3]);   // consi, i+3
                    values[ijac++] = -2*(x[i+1]-x[i+4]);   // consi, i+4
                    values[ijac++] = -2*(x[i+2]-x[i+5]);   // consi, i+5
                    consi++;                                
                    i+=3;
                    //g[consi++] = 
                    //    (x[i]-x[i+3])*(x[i]-x[i+3]) + (x[i+1]-x[i+4])*(x[i+1]-x[i+4]) + (x[i+2]-x[i+5])*(x[i+2]-x[i+5])                
                }
                i+=3;
            }        
        }
        
        assert (consi == num_sphere_constraints + num_break_constraints);
        if (num_clash_constraints > 0)
        {
            // the clash constraints            
            for (Index i=0; i < (num_loci-1)*3; i+=3)
            {
                for (Index j=i+3; j < num_loci*3; j+=3)
                {                
                    values[ijac++] = 2*(x[i]-x[j]);     // consi, i
                    values[ijac++] = 2*(x[i+1]-x[j+1]); // consi, i+1
                    values[ijac++] = 2*(x[i+2]-x[j+2]); // consi, i+2
                    
                    values[ijac++] = -2*(x[i]-x[j]);    // consi, j
                    values[ijac++] = -2*(x[i+1]-x[j+1]);    // consi, j+1
                    values[ijac++] = -2*(x[i+2]-x[j+2]);    // consi, j+2                
                                    
                    consi++;                                
                    //g[consi++] = (x[i]-x[j])*(x[i]-x[j]) + (x[i+1]-x[j+1])*(x[i+1]-x[j+1]) + (x[i+2]-x[j+2])*(x[i+2]-x[j+2]);
                }
            }
        }
            
        assert (consi == num_sphere_constraints + num_break_constraints + num_clash_constraints);
        
        if (num_dist_constraints > 0)
        {            
            INTERACTION_T *interaction = mygenome->get_first_interaction();
            while (interaction != NULL)
            {
            
                //printf ("Begin int values: ijac = %d, consi = %d\n", ijac, consi);
                int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
                int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);
                values[ijac++] = 2*(x[l]-x[r]);
                values[ijac++] = 2*(x[l+1]-x[r+1]);
                values[ijac++] = 2*(x[l+2]-x[r+2]);
                values[ijac++] = -2*(x[l]-x[r]);
                values[ijac++] = -2*(x[l+1]-x[r+1]);
                values[ijac++] = -2*(x[l+2]-x[r+2]);
                               
                
                //g[consi] = (x[l]-x[r])*(x[l]-x[r]) + 
                //          (x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                //          (x[l+2]-x[r+2])*(x[l+2]-x[r+2]) - x[xvari]*x[xvari];                
                
                interaction = interaction->next;
                consi++;
            }    
        }          
        
        //for (int i=1; i <= NUM_CHROMS; i++)
        {
            int i=12;
            int chr = mygenome->get_chrom_index(i);
            int centro_x = mygenome->get_x_index(chr, centromeres[chr]/bp_per_locus); 
                                    
            values[ijac++] = 2*(x[centro_x]-CENTRO_CENTER);
            values[ijac++] = 2*x[centro_x+1];
            values[ijac++] = 2*x[centro_x+2];
            consi++;
        }            
        
        for (Index i=0; i<num_small_sphere_constraints-1; i++)
        {
            int chr12 = mygenome->get_chrom_index(12);
            int centro_x = mygenome->get_x_index(chr12, RDNA_POSITION/bp_per_locus + i);
            values[ijac++] = 2*(x[centro_x]-NUCLEO_CENTER);
            values[ijac++] = 2*x[centro_x+1];
            values[ijac++] = 2*x[centro_x+2];
            consi++;                
        }           
        
        // if usedelta
        /*
        if (num_interactions > 0)
        {            
            INTERACTION_T *interaction = mygenome->get_first_interaction();
            while (interaction != NULL)
            {
            
                //printf ("Begin int values: ijac = %d, consi = %d\n", ijac, consi);
                int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
                int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);;
                values[ijac++] = 2*(x[l]-x[r]);
                values[ijac++] = 2*(x[l+1]-x[r+1]);
                values[ijac++] = 2*(x[l+2]-x[r+2]);
                values[ijac++] = -2*(x[l]-x[r]);
                values[ijac++] = -2*(x[l+1]-x[r+1]);
                values[ijac++] = -2*(x[l+2]-x[r+2]);
                values[ijac++] = 2 * x[xvari];
                               
                
                //g[consi] = (x[l]-x[r])*(x[l]-x[r]) + 
                //          (x[l+1]-x[r+1])*(x[l+1]-x[r+1]) + 
                //          (x[l+2]-x[r+2])*(x[l+2]-x[r+2]) - x[xvari]*x[xvari];                
                
                interaction = interaction->next;
                consi++;
                xvari++;
            }    
        }  
        */
        assert(consi == m);
    }
    
    //printf ("ijac=%d, nele_jac=%d\n", ijac, nele_jac);
    assert (ijac == nele_jac);
    assert (n == xvari);
    
    //printf ("Eval jac g OK\n");
    return true;
}

//return the structure or values of the hessian
bool genome_ipopt_nlp::eval_h(Index n, const Number* x, bool new_x,
                       Number obj_factor, Index m, const Number* lambda,
                       bool new_lambda, Index nele_hess, Index* iRow,
                       Index* jCol, Number* values)
{
    //printf ("Eval H OK\n");
    return false;   // if it's Quasi-Newton, just return false
}

void genome_ipopt_nlp::finalize_solution(SolverReturn status,
                                  Index n, const Number* x, const Number* z_L, const Number* z_U,
                                  Index m, const Number* g, const Number* lambda,
                                  Number obj_value,
				  const IpoptData* ip_data,
				  IpoptCalculatedQuantities* ip_cq)
{
  // here is where we would store the solution to variables, or write to a file, etc
  // so we could use the solution.

  // For this example, we write the solution to the console
//   printf("\n\nSolution of the primal variables, x\n");
//   for (Index i=0; i<n; i++) {
//     printf("x[%d] = %e\n", i, x[i]);
//   }

//   printf("\n\nSolution of the bound multipliers, z_L and z_U\n");
//   for (Index i=0; i<n; i++) {
//     printf("z_L[%d] = %e\n", i, z_L[i]);
//   }
//   for (Index i=0; i<n; i++) {
//     printf("z_U[%d] = %e\n", i, z_U[i]);
//   }

  printf("\n\nObjective value\n");
  printf("f(x*) = %e\n", obj_value);
  
  //printf("\nObjective value minus the forced distances: %e\n", 
  //  obj_value - num_interactions*min_dist*min_dist);

//   printf("\nFinal value of the constraints:\n");
//   for (Index i=0; i<m ;i++) {
//     printf("g(%d) = %e\n", i, g[i]);
//   }
  
    // now pass the info to mygenome and print pdb file
    mygenome->set_coords((double*)x);
    char output_filename[1000];
    sprintf (output_filename, "%s", output_pdb); // output_pdb already has .pdb
    mygenome->print_1D_3D_genome(output_filename); // Diament & Tuller, May 2014
    sprintf (output_filename, "%s.txt", output_pdb); 
    mygenome->save_txt(output_filename);
}



bool genome_ipopt_nlp::write_cplex_input (char *cplex_input_filename)
{
    FILE* output;
    if (open_file(cplex_input_filename, "w", TRUE, "CPLEX", "CPLEX", &output) == 0) {
        exit(1);
    }
    fprintf (output, "MINIMIZE\n");
        
    // first compute the quadratic matrix (qmatrix) of the objective function            
    
    int **qmatrix = new int* [num_variables];
    for (int i=0; i < num_variables; i++)
    {
        qmatrix[i] = new int[num_variables];
        for (int j=i; j < num_variables; j++)
        {
            qmatrix[i][j] = 0;
        }
    } 
    
    
    // First, I have to figure out what's the coefficient of every x_i * x_j pair of variables
    // Cplex doesn't like as input x1^2 + x1^2, is has to be 2 x1^2.
    if (num_interactions > 0)
    {                      
        INTERACTION_T *interaction = mygenome->get_first_interaction();
        while (interaction != NULL)
        {
            int l = mygenome->get_x_index(interaction->chrom1_index, interaction->locus1_index);
            int r = mygenome->get_x_index(interaction->chrom2_index, interaction->locus2_index);
            int f;
            if (use_weights == 1)  
            { 
                f = interaction->freq;
                printf ("NOT IMPLEMENTED YET\n");
                exit(1);
            }
            else
            {
                // make sure l < r;
                if (l > r)
                {
                    int temp = l;
                    l = r;
                    r = temp; 
                }
                qmatrix[l][l] += 1;         qmatrix[l][r] += -2;        qmatrix[r][r] += 1;
                qmatrix[l+1][l+1] += 1;     qmatrix[l+1][r+1] += -2;    qmatrix[r+1][r+1] += 1;
                qmatrix[l+2][l+2] += 1;     qmatrix[l+2][r+2] += -2;    qmatrix[r+2][r+2] += 1;                
            }                
            interaction = interaction->next;            
        }
    }            
        
    // add the break and clash lower bound constraints in the objective, so increment the corresponding coefficients
    // This didn't work, CPLEX gives the following message:
    //  Dual infeasible due to empty column 's0'.
    //      Presolve - Unbounded or infeasible.
    /*
    int i=0; 
    int svar = 0;
    if (num_break_constraints > 0)
    {
        for (int c=0; c < mygenome->get_num_chroms(); c++)
        {
            for (int l=0; l < mygenome->get_num_loci(c)-1; l++)
            {        
                //fprintf (output, "\n\tx%d^2 - 2 x%d*x%d + x%d^2 ", i, i, i+3, i+3);
                //fprintf (output, " + x%d^2 - 2 x%d*x%d + x%d^2 ", i+1, i+1, i+4, i+4);
                //fprintf (output, " + x%d^2 - 2 x%d*x%d + x%d^2 ", i+2, i+2, i+5, i+5);
                qmatrix[i][i] +=1;          qmatrix[i][i+3] += -2;          qmatrix[i+3][i+3] += 1;
                qmatrix[i+1][i+1] +=1;      qmatrix[i+1][i+4] += -2;        qmatrix[i+4][i+4] += 1;
                qmatrix[i+2][i+2] +=1;      qmatrix[i+2][i+5] += -2;        qmatrix[i+5][i+5] += 1;
                svar++;                      
                i+=3;
            }
            i+=3;
        }
    }
    
    // the clash constraints
    int cc=0;
    if (num_clash_constraints > 0)
    {
        int j;        
        for (i=0; i < (num_loci-1)*3; i+=3)
        {
            for (j=i+3; j < num_loci*3; j+=3)
            {
                //fprintf (output, "\n\tx%d^2 - 2 x%d*x%d + x%d^2 ", i, i, j, j);
                //fprintf (output, " + x%d^2 - 2 x%d*x%d + x%d^2 ", i+1, i+1, j+1, j+1);
                //fprintf (output, " + x%d^2 - 2 x%d*x%d + x%d^2 ] - s%d = 0\n", i+2, i+2, j+2, j+2, svar++);  
                qmatrix[i][i] +=1;          qmatrix[i][j] += -2;          qmatrix[j][j] += 1;
                qmatrix[i+1][i+1] +=1;      qmatrix[i+1][j+1] += -2;      qmatrix[j+1][j+1] += 1;
                qmatrix[i+2][i+2] +=1;      qmatrix[i+2][j+2] += -2;      qmatrix[j+2][j+2] += 1;
                svar++;                
            }
        }                        
    }           
    */
    
    // write the objective function
    fprintf (output, "\t[ ");    
    int first = 1;
    int this_row = 0;
    for (int i=0; i < num_variables; i++)
    {
        for (int j=i; j < num_variables; j++)
        {
            if (qmatrix[i][j] != 0)
            {
                fprintf (output, "%+d x%d * x%d ", qmatrix[i][j], i, j);
                this_row++;
                if (this_row == 10) 
                {
                    fprintf (output, "\n\t");
                    this_row = 0;
                } 
            }
        }
    }
    // NOTE: adding y0*y1 to the objective makes it not be positive semi-definite
    //fprintf (output, " + y0 * y1 ");
        
    
    fprintf (output, "\t ] / 2 \n\t");
    
    // Now add the slack variables s
    /*
    for (int i=0; i < svar; i++)
    {
        fprintf (output, " - s%d", i);
        if ((i+1)%10 == 0)  
            fprintf (output, "\n\t");
    } 
    */           
    
    // deallocate the space for qmatrix;
    for (int i=0; i < num_variables; i++)   delete [] qmatrix[i];
    delete [] qmatrix;
    
    fprintf (output, "\nSUBJECT TO\n");
    
    // NOW WRITE THE CONSTRAINTS
    // first the sphere constraints
    // 0 <= d^2(p_i,0) <= 1, for all loci p_i
    for (int i=0; i<num_sphere_constraints*3; i+=3)
    {
        // the lower bound doesn't make sense, it's true anyway
        //fprintf (output, "\tsc_l%d: [ x%d^2 + x%d^2 + x%d^2 ] >= 0\n", i/3, i, i+1, i+2);
        fprintf (output, "\tsc_u%d: [ x%d^2 + x%d^2 + x%d^2 ] <= %e\n", i/3, i, i+1, i+2, sphere_radius*sphere_radius);
    }
    
    // the break constraints   
    int i=0; 
    int bvar = 0;
    int yvar = 0;
    if (num_break_constraints > 0)
    {
        for (int c=0; c < mygenome->get_num_chroms(); c++)
        {
            for (int l=0; l < mygenome->get_num_loci(c)-1; l++)
            {        
                fprintf (output, "\n");                
                
                // NOTE: when I use >= with quadratic, the matrix is not positive semi-definite
                // I'm doing absolute values. For some reason, it didn't work to add 8 constraints, it says infeasible. I'm introducing new variables
//                 fprintf (output, "\tbc_l1_%d:  x%d - x%d + x%d - x%d + x%d - x%d >= %e\n", i/3, i, i+3, i+1, i+4, i+2, i+5, min_dist);
//                 fprintf (output, "\tbc_l2_%d:  x%d - x%d + x%d - x%d - x%d + x%d >= %e\n", i/3, i, i+3, i+1, i+4, i+2, i+5, min_dist);
//                 fprintf (output, "\tbc_l3_%d:  x%d - x%d - x%d + x%d + x%d - x%d >= %e\n", i/3, i, i+3, i+1, i+4, i+2, i+5, min_dist);
//                 fprintf (output, "\tbc_l4_%d:  x%d - x%d - x%d + x%d - x%d + x%d >= %e\n", i/3, i, i+3, i+1, i+4, i+2, i+5, min_dist);
//                 fprintf (output, "\tbc_l5_%d: -x%d + x%d + x%d - x%d + x%d - x%d >= %e\n", i/3, i, i+3, i+1, i+4, i+2, i+5, min_dist);
//                 fprintf (output, "\tbc_l6_%d: -x%d + x%d + x%d - x%d - x%d + x%d >= %e\n", i/3, i, i+3, i+1, i+4, i+2, i+5, min_dist);
//                 fprintf (output, "\tbc_l7_%d: -x%d + x%d - x%d + x%d + x%d - x%d >= %e\n", i/3, i, i+3, i+1, i+4, i+2, i+5, min_dist);
//                 fprintf (output, "\tbc_l8_%d: -x%d + x%d - x%d + x%d - x%d + x%d >= %e\n", i/3, i, i+3, i+1, i+4, i+2, i+5, min_dist); 
                                                
                // I tried introducing additional variables and constraints as follows, but it doesn't set y to the |x|, it just makes y big enough, and x very small
//                fprintf (output, "\tbc_l1_%d: y%d + y%d + y%d >= %e\n", i/3, 
//                    newvar, newvar+1, newvar+2, min_dist);
//                fprintf (output, "\tbc_l2_%d: y%d + x%d - x%d >= 0\n", i/3, newvar, i, i+3);
//                fprintf (output, "\tbc_l3_%d: y%d - x%d + x%d >= 0\n", i/3, newvar, i, i+3);
//                fprintf (output, "\tbc_l4_%d: y%d + x%d - x%d >= 0\n", i/3, newvar+1, i+1, i+4);
//                fprintf (output, "\tbc_l5_%d: y%d - x%d + x%d >= 0\n", i/3, newvar+1, i+1, i+4);
//                fprintf (output, "\tbc_l6_%d: y%d + x%d - x%d >= 0\n", i/3, newvar+2, i+2, i+5);
//                fprintf (output, "\tbc_l7_%d: y%d - x%d + x%d >= 0\n", i/3, newvar+2, i+2, i+5);                
//                newvar+=3;                         
                                
                // trying to do a different variable replacement
                // every absolute value is replaced by the difference between 2 non-negative numbers
                // this would only work if I could also add y0*y1=0, but I can't do that because it's not positive semi-definite
/*                fprintf (output, "\tbc_l1_%d: y%d + y%d + y%d + y%d + y%d + y%d >= %e\n", i/3, 
                    newvar, newvar+1, newvar+2, newvar+3, newvar+4, newvar+5, min_dist);
                fprintf (output, "\tbc_l2_%d: x%d - x%d - y%d + y%d = 0\n", i/3, i, i+3, newvar, newvar+1);
                fprintf (output, "\tbc_l3_%d: x%d - x%d - y%d + y%d = 0\n", i/3, i+1, i+4, newvar+2, newvar+3);
                fprintf (output, "\tbc_l4_%d: x%d - x%d - y%d + y%d = 0\n", i/3, i+2, i+5, newvar+4, newvar+5);
                newvar+=6;                         */
                
                // trying to do a different variable replacement
                // every x that needs to be inside absolute value is replaced by the difference between 2 non-negative numbers
                // the absolute value is replaced by the sum of these 2 numbers
                // add a binary variable
                // add constraints with the binary variable
                // this would only work if I could also add y0*y1=0, but I can't do that because it's not positive semi-definite
                fprintf (output, "\tbc_l1_%d: y%d + y%d + y%d + y%d + y%d + y%d >= %e\n", i/3, 
                    yvar, yvar+1, yvar+2, yvar+3, yvar+4, yvar+5, min_dist);
                fprintf (output, "\tbc_l2_%d: x%d - x%d - y%d + y%d = 0\n", i/3, i, i+3, yvar, yvar+1);
                fprintf (output, "\tbc_l3_%d: x%d - x%d - y%d + y%d = 0\n", i/3, i+1, i+4, yvar+2, yvar+3);
                fprintf (output, "\tbc_l4_%d: x%d - x%d - y%d + y%d = 0\n", i/3, i+2, i+5, yvar+4, yvar+5);                
                // add binary variables b
                fprintf (output, "\tbc_l5_%d: y%d - %e b%d <= 0\n",  i/3, yvar, sphere_radius, bvar);
                fprintf (output, "\tbc_l6_%d: y%d + %e b%d <= %e\n", i/3, yvar+1, sphere_radius, bvar, sphere_radius);
                fprintf (output, "\tbc_l7_%d: y%d - %e b%d <= 0\n",  i/3, yvar+2, sphere_radius, bvar+1);
                fprintf (output, "\tbc_l8_%d: y%d + %e b%d <= %e\n", i/3, yvar+3, sphere_radius, bvar+1, sphere_radius);
                fprintf (output, "\tbc_l9_%d: y%d - %e b%d <= 0\n",  i/3, yvar+4, sphere_radius, bvar+2);
                fprintf (output, "\tbc_l10_%d: y%d + %e b%d <= %e\n",i/3, yvar+5, sphere_radius, bvar+2, sphere_radius);
                yvar+=6;                                                         
                bvar+=3;
                                
                                
                // trying to add this in the objective                                
//                fprintf (output, "\tbc_l_%d: [ x%d^2 - 2 x%d*x%d + x%d^2 ", i/3, i, i, i+3, i+3);
//                fprintf (output, " + x%d^2 - 2 x%d*x%d + x%d^2 ", i+1, i+1, i+4, i+4);
//                fprintf (output, " + x%d^2 - 2 x%d*x%d + x%d^2 ] + s%d = 0\n", i+2, i+2, i+5, i+5, svar++);                  
                                
                fprintf (output, "\tbc_u_%d: [ x%d^2 - 2 x%d*x%d + x%d^2 ", i/3, i, i, i+3, i+3);
                fprintf (output, " + x%d^2 - 2 x%d*x%d + x%d^2 ", i+1, i+1, i+4, i+4);
                fprintf (output, " + x%d^2 - 2 x%d*x%d + x%d^2 ] <= %e\n", i+2, i+2, i+5, i+5, max_dist*max_dist);
                
//                fprintf (output, "\tbc_ineq_u%d: [ x%d^2 + x%d^2 + x%d^2 ] <= %e\n", 
//                    i/3, num_variables+i/3, num_variables+i/3+1, num_variables+i/3+2, max_dist*max_dist);
//                fprintf (output, "\tbc_eq1_%d:     x%d - x%d - x%d = 0\n", i/3, i, i+3, num_variables+i/3);
//                fprintf (output, "\tbc_eq2_%d:     x%d - x%d - x%d = 0\n", i/3, i+1, i+4, num_variables+i/3+1);
//                fprintf (output, "\tbc_eq3_%d:     x%d - x%d - x%d = 0\n", i/3, i+2, i+5, num_variables+i/3+2);
                    
                i+=3;
            }
            i+=3;
        }
    }
    
    
    /*
    // the clash constraints
    int cc=0;
    if (num_clash_constraints > 0)
    {
        int j;        
        for (i=0; i < (num_loci-1)*3; i+=3)
        {
            for (j=i+3; j < num_loci*3; j+=3)
            {
                fprintf (output, "\n");   
//                 // the same as above, use absolute values and introduce new variables
//                 fprintf (output, "\tcc_l1_%d: y%d + y%d + y%d >= %e\n", cc,  
//                     newvar, newvar+1, newvar+2, min_dist);
//                 fprintf (output, "\tcc_l2_%d: y%d + x%d - x%d >= 0\n", cc, newvar, i, j);
//                 fprintf (output, "\tcc_l3_%d: y%d - x%d + x%d >= 0\n", cc, newvar, i, j);
//                 fprintf (output, "\tcc_l4_%d: y%d + x%d - x%d >= 0\n", cc, newvar+1, i+1, j+1);
//                 fprintf (output, "\tcc_l5_%d: y%d - x%d + x%d >= 0\n", cc, newvar+1, i+1, j+1);
//                 fprintf (output, "\tcc_l6_%d: y%d + x%d - x%d >= 0\n", cc, newvar+2, i+2, j+2);
//                 fprintf (output, "\tcc_l7_%d: y%d - x%d + x%d >= 0\n", cc, newvar+2, i+2, j+2);                
//                 cc++;
//                 newvar+=3;
                
//                 fprintf (output, "\tcc_l1_%d: y%d + y%d + y%d + y%d + y%d + y%d >= %e\n", i/3, 
//                     newvar, newvar+1, newvar+2, newvar+3, newvar+4, newvar+5, min_dist);
//                 fprintf (output, "\tcc_l2_%d: x%d - x%d - y%d + y%d = 0\n", i/3, i, j, newvar, newvar+1);
//                 fprintf (output, "\tcc_l3_%d: x%d - x%d - y%d + y%d = 0\n", i/3, i+1, j+1, newvar+2, newvar+3);
//                 fprintf (output, "\tcc_l4_%d: x%d - x%d - y%d + y%d = 0\n", i/3, i+2, j+2, newvar+4, newvar+5);
//                 newvar += 6;
                
                fprintf (output, "\tcc_l_%d: [ x%d^2 - 2 x%d*x%d + x%d^2 ", i/3, i, i, j, j);
                fprintf (output, " + x%d^2 - 2 x%d*x%d + x%d^2 ", i+1, i+1, j+1, j+1);
                fprintf (output, " + x%d^2 - 2 x%d*x%d + x%d^2 ] + s%d = 0\n", i+2, i+2, j+2, j+2, svar++);                  
                
                
            }
        }                        
    }    
    */
    
    
    // NOW ADD THE BOUNDS ON THE VARIABLES
    fprintf (output, "\nBOUNDS\n");
    for (int i=0; i < num_variables; i++)
    {
        fprintf (output, "\t-%e <= x%d <= %e\n", sphere_radius, i, sphere_radius); 
    }
/*    for (int i=0; i < svar; i++)
    {
        fprintf (output, "\t%e <= s%d <= %e\n", min_dist*min_dist, i, sphere_radius*sphere_radius);
    }*/
    
    fprintf (output, "\nBINARY\n");
    for (int i=0; i < bvar; i++)
    {
        fprintf (output, "\tb%d\n", i);
    }
    
    fprintf (output, "\nEND\n");
    fclose(output);
}
